/* the first grid with inputs and buttons */

.head-grid {
  background-color: #eee;
  display: grid;
  grid-template-columns: 10em 8em 1fr;
  grid-template-areas:
    "head head head"
    "left-buttons nodes-label nodes-buttons"
    "left-buttons phones-label phones-buttons"
    "left-buttons pdus-label pdus-buttons"
    "left-buttons leases-label leases-buttons"
    ;
  grid-column-gap: 2px;
  grid-row-gap: 2px;
  justify-items: center;
  align-items: center;
}

/* the line with URL and status */
#top-line {
    display: flex;
    flex-direction: row;
}

#top-line>#url-label {
    font-size: 20px;
    width: 12em;
    padding-top: 8px;
}
#top-line>input {
    width: 20em;
}
#top-line>#connection-status {
    width: 100%;
}

/* the leaft area with connection management buttons */
#left-buttons {
    height: 100%;
    display: flex;
    flex-direction: column;
    justify-content: space-evenly;
    align-items: stretch;
    border-top: 1px solid #888;
    border-right: 2px solid #888;
}

#left-buttons>button {
    font-size: 20px;
    background-color: #ddd;
}
.category-buttons {
    display: flex;
}

/* actual inputs and buttons area */
.head-grid>div {
    font-size: 30px;
    text-align: center;
    width: 100%;
}

button {
    padding: 4px;
    font-size: 12px;
    border-radius: 5px;
}

.red {
    background-color: Tomato;
}
.green {
    background-color: Teal;
}
input {
    width: 100%;
    font-size: 18px;
    padding: 5px 15px;
}

/* -------------------- contents */
.body-grid {
    background-color: #e0e0e8;
    display: grid;
    grid-template-columns: 36% 17% 17% 30%;
    grid-template-areas:
       "nodes-contents phones-contents pdus-contents leases-contents";
    grid-column-gap: 2px;
    grid-row-gap: 2px;
    justify-items: center;
    align-items: start;
}


/* actual contents */
.body-grid>div {
    font-size: 30px;
    text-align: left;
    width: 100%;
    padding: 20px;
    border: .5px solid blue;
}


span.date {
    font-size: 12px;
}
div.contents>ul>li>ul>li {
    font-size: 15px;
}
span.record {
    margin: 3px;
    padding: 2px;
    border: 0.5px solid black;
    border-radius: 3px;
}

div.records {
    display: flex;
    flex-wrap: wrap;
}

.small {
    font-size: 50%;
}
